# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
# !/usr/bin/env python
"""
Provides "diff-like" comparison of images.

Currently relies on matplotlib for image processing so limited to PNG format.

"""

import argparse
import codecs
import contextlib
from glob import glob
import json
import os.path
import shutil
import sys
import tempfile
import warnings

# Force tephi.tests to use the ```tkagg``` backend by using the '-d'
# command-line argument as idiff is an interactive tool that requires a
# gui interface.
sys.argv.append("-d")
import tephi.tests  # noqa

from PIL import Image  # noqa
import filelock  # noqa
import imagehash  # noqa
import matplotlib.pyplot as plt  # noqa
import matplotlib.image as mimg  # noqa
import matplotlib.testing.compare as mcompare  # noqa
from matplotlib.testing.exceptions import ImageComparisonFailure  # noqa
import matplotlib.widgets as mwidget  # noqa
import numpy as np  # noqa
import requests  # noqa


_POSTFIX_DIFF = "-failed-diff.png"
_POSTFIX_JSON = os.path.join("results", "imagerepo.json")
_POSTFIX_LOCK = os.path.join("results", "imagerepo.lock")


@contextlib.contextmanager
def temp_png(suffix=""):
    if suffix:
        suffix = f"-{suffix}"
    suffix = f"{suffix}.png"

    fd, fname = tempfile.mkstemp(suffix)
    os.close(fd)

    try:
        yield fname
    finally:
        os.remove(fname)


def diff_viewer(
    repo,
    key,
    repo_fname,
    phash,
    status,
    expected_fname,
    result_fname,
    diff_fname,
):
    fig = plt.figure(figsize=(14, 12))
    plt.suptitle(os.path.basename(expected_fname))
    ax = plt.subplot(221)
    ax.imshow(mimg.imread(expected_fname))
    ax = plt.subplot(222, sharex=ax, sharey=ax)
    ax.imshow(mimg.imread(result_fname))
    ax = plt.subplot(223, sharex=ax, sharey=ax)
    ax.imshow(mimg.imread(diff_fname))

    result_dir = os.path.dirname(result_fname)
    fname = f"{phash}.png"
    uri = os.path.join(tephi.tests.BASE_URL, fname)
    phash_fname = os.path.join(result_dir, fname)

    def accept(event):
        if uri not in repo[key]:
            # Ensure to maintain strict time order where the first uri
            # associated with the repo key is the oldest, and the last
            # uri is the youngest
            repo[key].append(uri)
            # Update the image repo.
            with open(repo_fname, "wb") as fo:
                json.dump(
                    repo,
                    codecs.getwriter("utf-8")(fo),
                    indent=4,
                    sort_keys=True,
                )
            os.rename(result_fname, phash_fname)
            msg = "ACCEPTED:  {} -> {}"
            print(
                msg.format(
                    os.path.basename(result_fname),
                    os.path.basename(phash_fname),
                )
            )
        else:
            msg = "DUPLICATE: {} -> {} (ignored)"
            print(
                msg.format(
                    os.path.basename(result_fname),
                    os.path.basename(phash_fname),
                )
            )
            os.remove(result_fname)
        os.remove(diff_fname)
        plt.close()

    def reject(event):
        if uri not in repo[key]:
            print("REJECTED:  {}".format(os.path.basename(result_fname)))
        else:
            msg = "DUPLICATE: {} -> {} (ignored)"
            print(
                msg.format(
                    os.path.basename(result_fname),
                    os.path.basename(phash_fname),
                )
            )
        os.remove(result_fname)
        os.remove(diff_fname)
        plt.close()

    def skip(event):
        # Let's keep both the result and the diff files.
        print("SKIPPED:   {}".format(os.path.basename(result_fname)))
        plt.close()

    ax_accept = plt.axes([0.59, 0.05, 0.1, 0.075])
    ax_reject = plt.axes([0.7, 0.05, 0.1, 0.075])
    ax_skip = plt.axes([0.81, 0.05, 0.1, 0.075])
    baccept = mwidget.Button(ax_accept, "Accept")
    baccept.on_clicked(accept)
    breject = mwidget.Button(ax_reject, "Reject")
    breject.on_clicked(reject)
    bskip = mwidget.Button(ax_skip, "Skip")
    bskip.on_clicked(skip)
    plt.text(0.59, 0.15, status, transform=fig.transFigure)
    plt.show()


def _calculate_hit(uris, phash, action):
    # Extract the hex basename strings from the uris.
    hexes = [os.path.splitext(os.path.basename(uri))[0] for uri in uris]
    # Create the expected perceptual image hashes from the uris.
    to_hash = imagehash.hex_to_hash
    expected = [to_hash(uri_hex) for uri_hex in hexes]
    # Calculate the hamming distance vector for the result hash.
    distances = [e - phash for e in expected]

    if action == "first":
        index = 0
    elif action == "last":
        index = -1
    elif action == "similar":
        index = np.argmin(distances)
    elif action == "different":
        index = np.argmax(distances)
    else:
        emsg = "Unknown action: {!r}"
        raise ValueError(emsg.format(action))

    return index, distances[index]


def step_over_diffs(result_dir, action, display=True):
    processed = False
    dname = os.path.dirname(tephi.tests.__file__)
    lock = filelock.FileLock(os.path.join(dname, _POSTFIX_LOCK))
    if action in ["first", "last"]:
        kind = action
    elif action in ["similar", "different"]:
        kind = "most {}".format(action)
    else:
        emsg = "Unknown action: {!r}"
        raise ValueError(emsg.format(action))
    if display:
        msg = (
            "\nComparing the {!r} expected image with "
            "the test result image."
        )
        print(msg.format(kind))

    # Remove old image diff results.
    target = os.path.join(result_dir, f"*{_POSTFIX_DIFF}")
    for fname in glob(target):
        os.remove(fname)

    with lock.acquire(timeout=30):
        # Load the imagerepo.
        repo_fname = os.path.join(dname, _POSTFIX_JSON)
        with open(repo_fname, "rb") as fi:
            repo = json.load(codecs.getreader("utf-8")(fi))

        # Filter out all non-test result image files.
        target_glob = os.path.join(result_dir, "result-*.png")
        results = []
        for fname in sorted(glob(target_glob)):
            # We only care about PNG images.
            try:
                im = Image.open(fname)
                if im.format != "PNG":
                    # Ignore - it's not a png image.
                    continue
            except IOError:
                # Ignore - it's not an image.
                continue
            results.append(fname)

        count = len(results)

        for count_index, result_fname in enumerate(results):
            key = os.path.splitext("-".join(result_fname.split("-")[1:]))[0]
            try:
                # Calculate the test result perceptual image hash.
                phash = imagehash.phash(
                    Image.open(result_fname), hash_size=tephi.tests._HASH_SIZE
                )
                uris = repo[key]
                hash_index, distance = _calculate_hit(uris, phash, action)
                uri = uris[hash_index]
            except KeyError:
                wmsg = "Ignoring unregistered test result {!r}."
                warnings.warn(wmsg.format(key))
                continue
            with temp_png(key) as expected_fname:
                processed = True
                response = requests.get(uri)
                if response.status_code == requests.codes.ok:
                    with open(expected_fname, "wb") as fo:
                        fo.write(response.content)
                else:
                    # Perhaps the uri has not been pushed into the repo yet,
                    # so check if a local "developer" copy is available ...
                    local_fname = os.path.join(
                        result_dir, os.path.basename(uri)
                    )
                    if not os.path.isfile(local_fname):
                        emsg = "Bad URI {!r} for test {!r}."
                        raise ValueError(emsg.format(uri, key))
                    else:
                        # The temporary expected filename has the test name
                        # baked into it, and is used in the diff plot title.
                        # So copy the local file to the exected file to
                        # maintain this helpfulness.
                        shutil.copy(local_fname, expected_fname)
                try:
                    mcompare.compare_images(
                        expected_fname, result_fname, tol=0
                    )
                except Exception as e:
                    if isinstance(e, ValueError) or isinstance(
                        e, ImageComparisonFailure
                    ):
                        print(
                            "Could not compare {}: {}".format(result_fname, e)
                        )
                        continue
                    else:
                        # Propagate the exception, keeping the stack trace
                        raise
                diff_fname = os.path.splitext(result_fname)[0] + _POSTFIX_DIFF
                args = expected_fname, result_fname, diff_fname
                if display:
                    msg = "Image {} of {}: hamming distance = {} " "[{!r}]"
                    status = msg.format(count_index + 1, count, distance, kind)
                    prefix = repo, key, repo_fname, phash, status
                    yield prefix + args
                else:
                    yield args
        if display and not processed:
            print("\nThere are no tephi test result images to process.\n")


if __name__ == "__main__":
    default = os.path.join(
        os.path.dirname(tephi.tests.__file__), "result_image_comparison"
    )
    description = "Tephi graphic test difference tool."
    formatter_class = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(
        description=description, formatter_class=formatter_class
    )
    help = "path to tephi tests result image directory (default: %(default)s)"
    parser.add_argument("--resultdir", "-r", default=default, help=help)
    help = (
        'force "tephi.tests" to use the tkagg backend (default: %(default)s)'
    )
    parser.add_argument("-d", action="store_true", default=True, help=help)
    help = """
first     - compare result image with first (oldest) expected image
last      - compare result image with last (youngest) expected image
similar   - compare result image with most similar expected image (default)
different - compare result image with most unsimilar expected image
"""
    choices = ("first", "last", "similar", "different")
    parser.add_argument(
        "action", nargs="?", choices=choices, default="similar", help=help
    )
    args = parser.parse_args()
    result_dir = args.resultdir
    if not os.path.isdir(result_dir):
        emsg = "Invalid results directory: {}"
        raise ValueError(emsg.format(result_dir))
    for args in step_over_diffs(result_dir, args.action):
        diff_viewer(*args)



# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
The tephi module provides tephigram plotting of pressure, temperature and wind
barb data.

.. warning::
    This is a beta release module and is liable to change.

"""
from collections import namedtuple
from collections.abc import Iterable
from functools import partial
from matplotlib.font_manager import FontProperties
import matplotlib.pyplot as plt
from mpl_toolkits.axisartist.grid_helper_curvelinear import (
    GridHelperCurveLinear,
)
from mpl_toolkits.axisartist import Subplot
import numbers
import numpy as np
import os.path

from . import isopleths
from . import transforms


__version__ = "0.2.4"


#
# Miscellaneous constants.
#
DEFAULT_WIDTH = 700  # in pixels

ISOBAR_SPEC = [(25, 0.03), (50, 0.10), (100, 0.25), (200, 1.5)]
ISOBAR_LINE = {"color": "blue", "linewidth": 0.5, "clip_on": True}
ISOBAR_TEXT = {
    "size": 8,
    "color": "blue",
    "clip_on": True,
    "va": "bottom",
    "ha": "right",
}
ISOBAR_FIXED = [50, 1000]

WET_ADIABAT_SPEC = [(1, 0.05), (2, 0.15), (4, 1.5)]
WET_ADIABAT_LINE = {"color": "orange", "linewidth": 0.5, "clip_on": True}
WET_ADIABAT_TEXT = {
    "size": 8,
    "color": "orange",
    "clip_on": True,
    "va": "bottom",
    "ha": "left",
}
WET_ADIABAT_FIXED = None

MIXING_RATIO_SPEC = [(1, 0.05), (2, 0.18), (4, 0.3), (8, 1.5)]
MIXING_RATIO_LINE = {"color": "green", "linewidth": 0.5, "clip_on": True}
MIXING_RATIO_TEXT = {
    "size": 8,
    "color": "green",
    "clip_on": True,
    "va": "bottom",
    "ha": "right",
}
MIXING_RATIOS = [
    0.001,
    0.002,
    0.005,
    0.01,
    0.02,
    0.03,
    0.05,
    0.1,
    0.15,
    0.2,
    0.3,
    0.4,
    0.5,
    0.6,
    0.8,
    1.0,
    1.5,
    2.0,
    2.5,
    3.0,
    4.0,
    5.0,
    6.0,
    7.0,
    8.0,
    9.0,
    10.0,
    12.0,
    14.0,
    16.0,
    18.0,
    20.0,
    24.0,
    28.0,
    32.0,
    36.0,
    40.0,
    44.0,
    48.0,
    52.0,
    56.0,
    60.0,
    68.0,
    80.0,
]
MIXING_RATIO_FIXED = None

MIN_PRESSURE = 50  # mb = hPa
MAX_PRESSURE = 1000  # mb = hPa
MIN_THETA = 0  # degC
MAX_THETA = 250  # degC
MIN_WET_ADIABAT = 1  # degC
MAX_WET_ADIABAT = 60  # degC
MIN_TEMPERATURE = -50  # degC


RESOURCES_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), "etc")
DATA_DIR = os.path.join(RESOURCES_DIR, "test_data")


def loadtxt(*filenames, **kwargs):
    """
    Load one or more text files of pressure, temperature, wind speed and wind
    direction value sets.

    Each line should contain, at minimum, a single pressure value (mb or hPa),
    and a single temperature value (degC), but may also contain a dewpoint
    value (degC), wind speed (knots) and wind direction value (degrees from
    north).

    Note that blank lines and comment lines beginning with a '#' are ignored.

    For example:

    >>> import os.path
    >>> import tephi

    >>> winds = os.path.join(tephi.DATA_DIR, 'barbs.txt')
    >>> columns = ('pressure', 'dewpoint', 'wind_speed', 'wind_direction')
    >>> data = tephi.loadtxt(winds, column_titles=columns)
    >>> pressure = data.pressure
    >>> dews = data.dewpoint
    >>> wind_speed = data.wind_speed
    >>> wind_direction = data.wind_direction

    .. seealso:: :func:`numpy.loadtxt`.

    Args:

    * filenames: one or more filenames.

    Kwargs:

    * column_titles:
        List of iterables, or None. If specified, should contain one title
        string for each column of data per specified file. If all of multiple
        files loaded have the same column titles, then only one tuple of column
        titles need be specified.

    * delimiter:
        The string used to separate values. This is passed directly to
        :func:`np.loadtxt`, which defaults to using any whitespace as delimiter
        if this keyword is not specified.

    * dtype:
        The datatype to cast the data in the text file to. Passed directly to
        :func:`np.loadtxt`.

    Returns:
        A :func:`collections.namedtuple` instance containing one tuple, named
        with the relevant column title if specified, for each column of data
        in the text file loaded. If more than one file is loaded, a sequence
        of namedtuples is returned.

    """

    def _repr(nt):
        """An improved representation of namedtuples over the default."""

        typename = nt.__class__.__name__
        fields = nt._fields
        n_fields = len(fields)
        return_str = "{}(\n".format(typename)
        for i, t in enumerate(fields):
            gap = " " * 4
            if i == n_fields - 1:
                ender = ""
            else:
                ender = "\n"
            return_str += "{}{}={!r}{}".format(gap, t, getattr(nt, t), ender)
        return_str += ")"
        return return_str

    column_titles = kwargs.pop("column_titles", None)
    delimiter = kwargs.pop("delimiter", None)
    dtype = kwargs.pop("dtype", "f4")

    if column_titles is not None:
        fields = column_titles[0]
        if not isinstance(column_titles, str):
            if isinstance(fields, Iterable) and not isinstance(fields, str):
                # We've an iterable of iterables - multiple titles is True.
                multiple_titles = True
                if len(column_titles) > len(filenames):
                    msg = "Received {} files but {} sets of column titles."
                    raise ValueError(
                        msg.format(len(column_titles), len(filenames))
                    )
            elif isinstance(fields, str):
                # We've an iterable of title strings - use for namedtuple.
                tephidata = namedtuple("tephidata", column_titles)
                multiple_titles = False
            else:
                # Whatever we've got it isn't iterable, so raise TypeError.
                msg = "Expected title to be string, got {!r}."
                raise TypeError(msg.format(type(column_titles)))
        else:
            msg = "Expected column_titles to be iterable, got {!r}."
            raise TypeError(msg.format(type(column_titles)))

    else:
        tephidata = namedtuple("tephidata", ("pressure", "temperature"))
        multiple_titles = False

    data = []
    for ct, arg in enumerate(filenames):
        if isinstance(arg, str):
            if os.path.isfile(arg):
                if multiple_titles:
                    tephidata = namedtuple("tephidata", column_titles[ct])
                tephidata.__repr__ = _repr
                payload = np.loadtxt(arg, dtype=dtype, delimiter=delimiter)
                item = tephidata(*payload.T)
                data.append(item)
            else:
                msg = "Item {} is either not a file or does not exist."
                raise OSError(msg.format(arg))

    if len(data) == 1:
        data = data[0]

    return data


class _FormatterTheta:
    """Dry adiabats potential temperature axis tick formatter."""

    def __call__(self, direction, factor, values):
        return [r"$\theta={:.1f}$".format(value) for value in values]


class _FormatterIsotherm:
    """Isotherms temperature axis tick formatter."""

    def __call__(self, direction, factor, values):
        return [r"  $T={:.1f}$".format(value) for value in values]


class Locator:
    """
    Determine the fixed step axis tick locations when called with a tick range.

    """

    def __init__(self, step):
        """
        Set the fixed step value for the axis tick locations.

        Generate tick location specification when called with a tick range.

        For example:

            >>> from tephi import Locator
            >>> locator = Locator(10)
            >>> locator(-45, 23)
            (array([-50, -40, -30, -20, -10,   0,  10,  20]), 8, 1)

        Args:

        * step: the step value for each axis tick.

        """
        self.step = int(step)

    def __call__(self, start, stop):
        """Calculate the axis ticks given the provided tick range."""

        step = self.step
        start = (int(start) / step) * step
        stop = (int(stop) / step) * step
        ticks = np.arange(start, stop + step, step)

        return ticks, len(ticks), 1


def _refresh_isopleths(axes):
    """
    Refresh the plot isobars, wet adiabats and mixing ratios and associated
    text labels.

    Args:

    * axes:
        Tephigram plotting :class:`matplotlib.axes.AxesSubplot` instance.

    Returns:
        Boolean, whether the plot has changed.

    """
    changed = False

    # Determine the current zoom level.
    xlim = axes.get_xlim()
    delta_xlim = xlim[1] - xlim[0]
    ylim = axes.get_ylim()
    zoom = delta_xlim / axes.tephigram_original_delta_xlim

    # Determine the display mid-point.
    x_point = xlim[0] + delta_xlim * 0.5
    y_point = ylim[0] + (ylim[1] - ylim[0]) * 0.5
    xy = np.array([[x_point, y_point]])
    xy_point = axes.tephigram_inverse.transform(xy)[0]

    for profile in axes.tephigram_profiles:
        profile.refresh()

    for isopleth in axes.tephigram_isopleths:
        changed = isopleth.refresh(zoom, xy_point) or changed

    return changed


def _handler(event):
    """Matplotlib event handler."""

    for axes in event.canvas.figure.axes:
        if hasattr(axes, "tephigram"):
            if _refresh_isopleths(axes):
                event.canvas.figure.show()


class _PlotGroup(dict):
    """
    Container for a related group of tephigram isopleths.

    Manages the creation and plotting of all isopleths within the group.

    """

    def __init__(
        self,
        axes,
        plot_func,
        text_kwargs,
        step,
        zoom,
        tags,
        fixed=None,
        xfocus=None,
    ):
        self.axes = axes
        self.text_kwargs = text_kwargs
        self.step = step
        self.zoom = zoom

        pairs = []
        for tag in tags:
            text = plt.text(0, 0, str(tag), **text_kwargs)
            text.set_bbox(
                dict(
                    boxstyle="Round,pad=0.3",
                    facecolor="white",
                    edgecolor="white",
                    alpha=0.5,
                    clip_on=True,
                    clip_box=self.axes.bbox,
                )
            )
            pairs.append((tag, [plot_func(tag), text]))

        dict.__init__(self, pairs)
        for line, text in self.values():
            line.set_visible(True)
            text.set_visible(True)
        self._visible = True

        if fixed is None:
            fixed = []

        if not isinstance(fixed, Iterable):
            fixed = [fixed]

        if zoom is None:
            self.fixed = set(tags)
        else:
            self.fixed = set(tags) & set(fixed)

        self.xfocus = xfocus

    def __setitem__(self, tag, item):
        emsg = "Cannot add or set an item into the plot group {!r}"
        raise ValueError(emsg.format(self.step))

    def __getitem__(self, tag):
        if tag not in self.keys():
            emsg = "Tag item {!r} is not a member of the plot group {!r}"
            raise KeyError(emsg.format(tag, self.step))
        return dict.__getitem__(self, tag)

    def refresh(self, zoom, xy_point):
        """
        Refresh all isopleths within the plot group.

        Args:

        * zoom:
            Zoom level of the current plot, relative to the initial plot.
        * xy_point:
            The center point of the current point, transformed into
            temperature and potential temperature.

        Returns:
            Boolean, whether the plot group has changed.

        """
        if self.zoom is None or zoom <= self.zoom:
            changed = self._item_on()
        else:
            changed = self._item_off()
        self._refresh_text(xy_point)
        return changed

    def _item_on(self, zoom=None):
        changed = False
        if zoom is None or self.zoom is None or zoom <= self.zoom:
            if not self._visible:
                for line, text in self.values():
                    line.set_visible(True)
                    text.set_visible(True)
                changed = True
                self._visible = True
        return changed

    def _item_off(self, zoom=None):
        changed = False
        if self.zoom is not None and (zoom is None or zoom > self.zoom):
            if self._visible:
                for tag, (line, text) in self.items():
                    if tag not in self.fixed:
                        line.set_visible(False)
                        text.set_visible(False)
                        changed = True
                        self._visible = False
        return changed

    def _generate_text(self, tag, xy_point):
        line, text = self[tag]
        x_data = line.get_xdata()
        y_data = line.get_ydata()

        if self.xfocus:
            delta = np.power(x_data - xy_point[0], 2)
        else:
            delta = np.power(x_data - xy_point[0], 2) + np.power(
                y_data - xy_point[1], 2
            )
        index = np.argmin(delta)
        text.set_position((x_data[index], y_data[index]))

    def _refresh_text(self, xy_point):
        if self._visible:
            for tag in self:
                self._generate_text(tag, xy_point)
        elif self.fixed:
            for tag in self.fixed:
                self._generate_text(tag, xy_point)


class _PlotCollection:
    """
    Container for tephigram isopleths.

    Manages the creation and plotting of all tephigram isobars, mixing ratio
    lines and pseudo saturated wet adiabats.

    """

    def __init__(
        self,
        axes,
        spec,
        stop,
        plot_func,
        text_kwargs,
        fixed=None,
        minimum=None,
        xfocus=None,
    ):
        if isinstance(stop, Iterable):
            if minimum and minimum > max(stop):
                emsg = "Minimum value of {!r} exceeds all other values"
                raise ValueError(emsg.format(minimum))

            items = [
                [step, zoom, set(stop[step - 1 :: step])]
                for step, zoom in sorted(spec, reverse=True)
            ]
        else:
            if minimum and minimum > stop:
                emsg = "Minimum value of {!r} exceeds maximum threshold {!r}"
                raise ValueError(emsg.format(minimum, stop))

            items = [
                [step, zoom, set(range(step, stop + step, step))]
                for step, zoom in sorted(spec, reverse=True)
            ]

        for index, item in enumerate(items):
            if minimum:
                item[2] = set([value for value in item[2] if value >= minimum])

            for subitem in items[index + 1 :]:
                subitem[2] -= item[2]

        self.groups = {
            item[0]: _PlotGroup(
                axes, plot_func, text_kwargs, *item, fixed=fixed, xfocus=xfocus
            )
            for item in items
            if item[2]
        }

        if not self.groups:
            emsg = "The plot collection failed to generate any plot groups"
            raise ValueError(emsg)

    def refresh(self, zoom, xy_point):
        """
        Refresh all isopleth groups within the plot collection.

        Args:

        * zoom:
            Zoom level of the current plot, relative to the initial plot.
        * xy_point:
            The center point of the current plot, transformed into
            temperature and potential temperature.

        Returns:
            Boolean, whether any plot group has changed.

        """
        changed = False

        for group in self.groups.values():
            changed = group.refresh(zoom, xy_point) or changed

        return changed


class Tephigram:
    """
    Generate a tephigram of one or more pressure and temperature data sets.

    """

    def __init__(
        self,
        figure=None,
        isotherm_locator=None,
        dry_adiabat_locator=None,
        anchor=None,
    ):
        """
        Initialise the tephigram transformation and plot axes.

        Kwargs:

        * figure:
            An existing :class:`matplotlib.figure.Figure` instance for the
            tephigram plot. If a figure is not provided, a new figure will
            be created by default.
        * isotherm_locator:
            A :class:`tephi.Locator` instance or a numeric step size
            for the isotherm lines.
        * dry_adiabat_locator:
            A :class:`tephi.Locator` instance or a numeric step size
            for the dry adiabat lines.
        * anchor:
            A sequence of two pressure, temperature pairs specifying the extent
            of the tephigram plot in terms of the bottom left hand corner and
            the top right hand corner. Pressure data points must be in units of
            mb or hPa, and temperature data points must be in units of degC.

        For example:

        .. plot::
            :include-source:

            import matplotlib.pyplot as plt
            from numpy import column_stack
            import os.path
            import tephi
            from tephi import Tephigram

            dew_point = os.path.join(tephi.DATA_DIR, 'dews.txt')
            dry_bulb = os.path.join(tephi.DATA_DIR, 'temps.txt')
            dew_data, temp_data = tephi.loadtxt(dew_point, dry_bulb)
            dews = column_stack((dew_data.pressure, dew_data.temperature))
            temps = column_stack((temp_data.pressure, temp_data.temperature))
            tpg = Tephigram()
            tpg.plot(dews, label='Dew-point', color='blue', linewidth=2)
            tpg.plot(temps, label='Dry-bulb', color='red', linewidth=2)
            plt.show()

        """
        if not figure:
            # Create a default figure.
            self.figure = plt.figure(0, figsize=(9, 9))
        else:
            self.figure = figure

        # Configure the locators.
        if isotherm_locator and not isinstance(isotherm_locator, Locator):
            if not isinstance(isotherm_locator, numbers.Number):
                raise ValueError("Invalid isotherm locator")
            locator_isotherm = Locator(isotherm_locator)
        else:
            locator_isotherm = isotherm_locator

        if dry_adiabat_locator and not isinstance(
            dry_adiabat_locator, Locator
        ):
            if not isinstance(dry_adiabat_locator, numbers.Number):
                raise ValueError("Invalid dry adiabat locator")
            locator_theta = Locator(dry_adiabat_locator)
        else:
            locator_theta = dry_adiabat_locator

        # Define the tephigram coordinate-system transformation.
        self.tephi_transform = transforms.TephiTransform()
        ghelper = GridHelperCurveLinear(
            self.tephi_transform,
            tick_formatter1=_FormatterIsotherm(),
            grid_locator1=locator_isotherm,
            tick_formatter2=_FormatterTheta(),
            grid_locator2=locator_theta,
        )
        self.axes = Subplot(self.figure, 1, 1, 1, grid_helper=ghelper)
        self.transform = self.tephi_transform + self.axes.transData
        self.axes.axis["isotherm"] = self.axes.new_floating_axis(1, 0)
        self.axes.axis["theta"] = self.axes.new_floating_axis(0, 0)
        self.axes.axis["left"].get_helper().nth_coord_ticks = 0
        self.axes.axis["left"].toggle(all=True)
        self.axes.axis["bottom"].get_helper().nth_coord_ticks = 1
        self.axes.axis["bottom"].toggle(all=True)
        self.axes.axis["top"].get_helper().nth_coord_ticks = 0
        self.axes.axis["top"].toggle(all=False)
        self.axes.axis["right"].get_helper().nth_coord_ticks = 1
        self.axes.axis["right"].toggle(all=True)
        self.axes.gridlines.set_linestyle("solid")

        self.figure.add_subplot(self.axes)

        # Configure default axes.
        axis = self.axes.axis["left"]
        axis.major_ticklabels.set_fontsize(10)
        axis.major_ticklabels.set_va("baseline")
        axis.major_ticklabels.set_rotation(135)
        axis = self.axes.axis["right"]
        axis.major_ticklabels.set_fontsize(10)
        axis.major_ticklabels.set_va("baseline")
        axis.major_ticklabels.set_rotation(-135)
        self.axes.axis["top"].major_ticklabels.set_fontsize(10)
        axis = self.axes.axis["bottom"]
        axis.major_ticklabels.set_fontsize(10)
        axis.major_ticklabels.set_ha("left")
        axis.major_ticklabels.set_va("top")
        axis.major_ticklabels.set_rotation(-45)

        # Isotherms: lines of constant temperature (degC).
        axis = self.axes.axis["isotherm"]
        axis.set_axis_direction("right")
        axis.set_axislabel_direction("-")
        axis.major_ticklabels.set_rotation(90)
        axis.major_ticklabels.set_fontsize(10)
        axis.major_ticklabels.set_va("bottom")
        axis.major_ticklabels.set_color("grey")
        axis.major_ticklabels.set_visible(False)  # turned-off

        # Dry adiabats: lines of constant potential temperature (degC).
        axis = self.axes.axis["theta"]
        axis.set_axis_direction("right")
        axis.set_axislabel_direction("+")
        axis.major_ticklabels.set_fontsize(10)
        axis.major_ticklabels.set_va("bottom")
        axis.major_ticklabels.set_color("grey")
        axis.major_ticklabels.set_visible(False)  # turned-off
        axis.line.set_linewidth(3)
        axis.line.set_linestyle("--")

        # Lock down the aspect ratio.
        self.axes.set_aspect(1.0)
        self.axes.grid(True)

        # Initialise the text formatter for the navigation status bar.
        self.axes.format_coord = self._status_bar

        # Factor in the tephigram transform.
        ISOBAR_TEXT["transform"] = self.transform
        WET_ADIABAT_TEXT["transform"] = self.transform
        MIXING_RATIO_TEXT["transform"] = self.transform

        # Create plot collections for the tephigram isopleths.
        func = partial(
            isopleths.isobar,
            MIN_THETA,
            MAX_THETA,
            self.axes,
            self.transform,
            ISOBAR_LINE,
        )
        self._isobars = _PlotCollection(
            self.axes,
            ISOBAR_SPEC,
            MAX_PRESSURE,
            func,
            ISOBAR_TEXT,
            fixed=ISOBAR_FIXED,
            minimum=MIN_PRESSURE,
        )

        func = partial(
            isopleths.wet_adiabat,
            MAX_PRESSURE,
            MIN_TEMPERATURE,
            self.axes,
            self.transform,
            WET_ADIABAT_LINE,
        )
        self._wet_adiabats = _PlotCollection(
            self.axes,
            WET_ADIABAT_SPEC,
            MAX_WET_ADIABAT,
            func,
            WET_ADIABAT_TEXT,
            fixed=WET_ADIABAT_FIXED,
            minimum=MIN_WET_ADIABAT,
            xfocus=True,
        )

        func = partial(
            isopleths.mixing_ratio,
            MIN_PRESSURE,
            MAX_PRESSURE,
            self.axes,
            self.transform,
            MIXING_RATIO_LINE,
        )
        self._mixing_ratios = _PlotCollection(
            self.axes,
            MIXING_RATIO_SPEC,
            MIXING_RATIOS,
            func,
            MIXING_RATIO_TEXT,
            fixed=MIXING_RATIO_FIXED,
        )

        # Initialise for the tephigram plot event handler.
        plt.connect("motion_notify_event", _handler)
        self.axes.tephigram = True
        self.axes.tephigram_original_delta_xlim = DEFAULT_WIDTH
        self.original_delta_xlim = DEFAULT_WIDTH
        self.axes.tephigram_transform = self.tephi_transform
        self.axes.tephigram_inverse = self.tephi_transform.inverted()
        self.axes.tephigram_isopleths = [
            self._isobars,
            self._wet_adiabats,
            self._mixing_ratios,
        ]

        # The tephigram profiles.
        self._profiles = []
        self.axes.tephigram_profiles = self._profiles

        # Center the plot around the anchor extent.
        self._anchor = anchor
        if self._anchor is not None:
            self._anchor = np.asarray(anchor)
            if (
                self._anchor.ndim != 2
                or self._anchor.shape[-1] != 2
                or len(self._anchor) != 2
            ):
                msg = (
                    "Invalid anchor, expecting [(bottom-left-pressure, "
                    "bottom-left-temperature), (top-right-pressure, "
                    "top-right-temperature)]"
                )
                raise ValueError(msg)
            (
                (bottom_pressure, bottom_temp),
                (top_pressure, top_temp),
            ) = self._anchor

            if (bottom_pressure - top_pressure) < 0:
                raise ValueError("Invalid anchor pressure range")
            if (bottom_temp - top_temp) < 0:
                raise ValueError("Invalid anchor temperature range")

            self._anchor = isopleths.Profile(anchor, self.axes)
            self._anchor.plot(visible=False)
            xlim, ylim = self._calculate_extents()
            self.axes.set_xlim(xlim)
            self.axes.set_ylim(ylim)

    def plot(self, data, **kwargs):
        """
        Plot the environmental lapse rate profile of the pressure and
        temperature data points.

        The pressure and temperature data points are transformed into
        potential temperature and temperature data points before plotting.

        By default, the tephigram will automatically center the plot around
        all profiles.

        .. warning::
            Pressure data points must be in units of mb or hPa, and temperature
            data points must be in units of degC.

        Args:

        * data: pressure and temperature pair data points.

        .. note::
            All keyword arguments are passed through to
            :func:`matplotlib.pyplot.plot`.

        For example:

        .. plot::
            :include-source:

            import matplotlib.pyplot as plt
            from tephi import Tephigram

            tpg = Tephigram()
            data = [[1006, 26.4], [924, 20.3], [900, 19.8],
                    [850, 14.5], [800, 12.9], [755, 8.3]]
            profile = tpg.plot(data, color='red', linestyle='--',
                               linewidth=2, marker='o')
            barbs = [(10, 45, 900), (20, 60, 850), (25, 90, 800)]
            profile.barbs(barbs)
            plt.show()

        For associating wind barbs with an environmental lapse rate profile,
        see :meth:`~tephi.isopleths.Profile.barbs`.

        """
        profile = isopleths.Profile(data, self.axes)
        profile.plot(**kwargs)
        self._profiles.append(profile)

        # Center the tephigram plot around all the profiles.
        if self._anchor is None:
            xlim, ylim = self._calculate_extents(xfactor=0.25, yfactor=0.05)
            self.axes.set_xlim(xlim)
            self.axes.set_ylim(ylim)

        # Refresh the tephigram plot isopleths.
        _refresh_isopleths(self.axes)

        # Show the plot legend.
        if "label" in kwargs:
            font_properties = FontProperties(size="x-small")
            plt.legend(
                loc="upper left",
                fancybox=True,
                shadow=True,
                prop=font_properties,
            )

        return profile

    def _status_bar(self, x_point, y_point):
        """Generate text for the interactive backend navigation status bar."""

        temperature, theta = transforms.convert_xy2Tt(x_point, y_point)
        pressure, _ = transforms.convert_Tt2pT(temperature, theta)
        xlim = self.axes.get_xlim()
        zoom = (xlim[1] - xlim[0]) / self.original_delta_xlim
        msg = "T:{:.2f}, theta:{:.2f}, phi:{:.2f} (zoom:{:.3f})"
        text = msg.format(
            float(temperature), float(theta), float(pressure), zoom
        )

        return text

    def _calculate_extents(self, xfactor=None, yfactor=None):
        min_x = min_y = 1e10
        max_x = max_y = -1e-10
        profiles = self._profiles
        transform = self.tephi_transform.transform

        if self._anchor is not None:
            profiles = [self._anchor]

        for profile in profiles:
            temperature = profile.temperature.reshape(-1, 1)
            theta = profile.theta.reshape(-1, 1)
            xy_points = transform(np.concatenate((temperature, theta), axis=1))
            x_points = xy_points[:, 0]
            y_points = xy_points[:, 1]
            min_x = np.min([min_x, np.min(x_points)])
            min_y = np.min([min_y, np.min(y_points)])
            max_x = np.max([max_x, np.max(x_points)])
            max_y = np.max([max_y, np.max(y_points)])

        if xfactor is not None:
            delta_x = max_x - min_x
            min_x, max_x = min_x - xfactor * delta_x, max_x + xfactor * delta_x

        if yfactor is not None:
            delta_y = max_y - min_y
            min_y, max_y = min_y - yfactor * delta_y, max_y + yfactor * delta_y

        return ([min_x, max_x], [min_y, max_y])



# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tephigram isopleth support for generating and plotting tephigram lines,
environment profiles and barbs.

"""
import math
from matplotlib.collections import PathCollection
from matplotlib.path import Path
import matplotlib.pyplot as plt
import matplotlib.transforms as mtransforms
import numpy as np
from scipy.interpolate import interp1d

from ._constants import CONST_CP, CONST_L, CONST_KELVIN, CONST_RD, CONST_RV
from . import transforms


# Wind barb speed (knots) ranges used since 1 January 1955.
_BARB_BINS = np.arange(20) * 5 + 3
_BARB_GUTTER = 0.1
_BARB_DTYPE = np.dtype(
    dict(
        names=("speed", "angle", "pressure", "barb"),
        formats=("f4", "f4", "f4", np.object),
    )
)

#
# Reference: http://www-nwp/~hadaa/tephigram/tephi_plot.html
#


def mixing_ratio(
    min_pressure, max_pressure, axes, transform, kwargs, mixing_ratio_value
):
    """
    Generate and plot a humidity mixing ratio line.

    A line of constant saturation mixing ratio with respect to a
    plane water surface (g kg-1).

    Args:

    * min_pressure:
        Minumum pressure, in mb or hPa, for the mixing ratio line extent.

    * max_pressure:
        Maximum pressure, in mb or hPa, for the mixing ratio line extent.

    * axes:
        Tephigram plotting :class:`matplotlib.axes.AxesSubplot` instance.

    * transform:
        Tephigram plotting transformation
        :class:`matplotlib.transforms.CompositeGenericTransform` instance.

    * kwargs:
        Keyword arguments for the mixing ratio :class:`matplotlib.lines.Line2D`
        instance.

    * mixing_ratio_value:
        The mixing ratio value to be plotted.

    Returns:
        The mixing ratio :class:`matplotlib.lines.Line2D` instance.

    """
    pressures = np.linspace(min_pressure, max_pressure, 100)
    temps = transforms.convert_pw2T(pressures, mixing_ratio_value)
    _, thetas = transforms.convert_pT2Tt(pressures, temps)
    (line,) = axes.plot(temps, thetas, transform=transform, **kwargs)

    return line


def isobar(min_theta, max_theta, axes, transform, kwargs, pressure):
    """
    Generate and plot an isobar line.

    A line of constant pressure (mb).

    Args:

    * min_theta:
        Minimum potential temperature, in degC, for the isobar extent.

    * max_theta:
        Maximum potential temperature, in degC, for the isobar extent.

    * axes:
        Tephigram plotting :class:`matplotlib.axes.AxesSubplot` instance.

    * transform:
        Tephigram plotting transformation
        :class:`matplotlib.transforms.CompositeGenericTransform` instance.

    * kwargs:
        Keyword arguments for the isobar :class:`matplotlib.lines.Line2D`
        instance.

    * pressure:
        The isobar pressure value, in mb or hPa, to be plotted.

    Returns:
       The isobar :class:`matplotlib.lines.Line2D` instance.

    """
    steps = 100
    thetas = np.linspace(min_theta, max_theta, steps)
    _, temps = transforms.convert_pt2pT([pressure] * steps, thetas)
    (line,) = axes.plot(temps, thetas, transform=transform, **kwargs)

    return line


def _wet_adiabat_gradient(min_temperature, pressure, temperature, dp):
    """
    Calculate the wet adiabat change in pressure and temperature.

    Args:

    * min_temperature:
        Minimum potential temperature, in degC, for the wet adiabat line
        extent.

    * pressure:
        Pressure point value, in mb or hPa, from which to calculate the
        gradient difference.

    * temperature:
        Potential temperature point value, in degC, from which to calculate
        the gradient difference.

    * dp:
        The wet adiabat change in pressure, in mb or hPa, from which to
        calculate the gradient difference.

    Returns:
        The gradient change as a pressure, potential temperature value pair.

    """

    # TODO: Discover the meaning of the magic numbers.

    kelvin = temperature + CONST_KELVIN
    lsbc = (CONST_L / CONST_RV) * ((1.0 / CONST_KELVIN) - (1.0 / kelvin))
    rw = 6.11 * np.exp(lsbc) * (0.622 / pressure)
    lrwbt = (CONST_L * rw) / (CONST_RD * kelvin)
    nume = ((CONST_RD * kelvin) / (CONST_CP * pressure)) * (1.0 + lrwbt)
    deno = 1.0 + (lrwbt * ((0.622 * CONST_L) / (CONST_CP * kelvin)))
    gradi = nume / deno
    dt = dp * gradi

    if (temperature + dt) < min_temperature:
        dt = min_temperature - temperature
        dp = dt / gradi

    return dp, dt


def wet_adiabat(
    max_pressure, min_temperature, axes, transform, kwargs, temperature
):
    """
    Generate and plot a pseudo saturated wet adiabat line.

    A line of constant equivalent potential temperature for saturated
    air parcels (degC).

    Args:

    * max_pressure:
        Maximum pressure, in mb or hPa, for the wet adiabat line extent.

    * min_temperature:
        Minimum potential temperature, in degC, for the wet adiabat line
        extent.

    * axes:
        Tephigram plotting :class:`matplotlib.axes.AxesSubplot` instance.

    * transform:
        Tephigram plotting transformation
        :class:`matplotlib.transforms.CompositeGenericTransform` instance.

    * kwargs:
        Keyword arguments for the mixing ratio :class:`matplotlib.lines.Line2D`
        instance.

    * temperature:
        The wet adiabat value, in degC, to be plotted.

    Returns:
        The wet adiabat :class:`matplotlib.lines.Line2D` instance.

    """
    temps = [temperature]
    pressures = [max_pressure]
    dp = -5.0

    for i in range(200):
        dp, dt = _wet_adiabat_gradient(
            min_temperature, pressures[i], temps[i], dp
        )
        temps.append(temps[i] + dt)
        pressures.append(pressures[i] + dp)

    _, thetas = transforms.convert_pT2Tt(pressures, temps)
    (line,) = axes.plot(temps, thetas, transform=transform, **kwargs)

    return line


class Barbs(object):
    """Generate a wind arrow barb."""

    def __init__(self, axes):
        """
        Create a wind arrow barb for the given axes.

        Args:

        * axes:
            A :class:`matplotlib.axes.AxesSubplot` instance.

        """
        self.axes = axes
        self.barbs = None
        self._gutter = None
        self._transform = axes.tephigram_transform + axes.transData
        self._kwargs = None
        self._custom_kwargs = None
        self._custom = dict(
            color=["barbcolor", "color", "edgecolor", "facecolor"],
            linewidth=["lw", "linewidth"],
            linestyle=["ls", "linestyle"],
        )

    @staticmethod
    def _uv(magnitude, angle):
        """
        Convert magnitude and angle measured in degrees to u and v components,
        where u is -x and v is -y.

        """
        angle = angle % 360
        u = v = 0
        # Snap the magnitude of the barb vector to fall into one of the
        # _BARB_BINS ensuring it's a multiple of five. Five is the increment
        # step size for decorating with barb with flags.
        magnitude = np.searchsorted(_BARB_BINS, magnitude, side="right") * 5
        modulus = angle % 90
        if modulus:
            quadrant = int(angle / 90)
            radians = math.radians(modulus)
            y = math.cos(radians) * magnitude
            x = math.sin(radians) * magnitude
            if quadrant == 0:
                u, v = -x, -y
            elif quadrant == 1:
                u, v = -y, x
            elif quadrant == 2:
                u, v = x, y
            else:
                u, v = y, -x
        else:
            angle = int(angle)
            if angle == 0:
                v = -magnitude
            elif angle == 90:
                u = -magnitude
            elif angle == 180:
                v = magnitude
            else:
                u = magnitude
        return u, v

    def _make_barb(self, temperature, theta, speed, angle):
        """Add the barb to the plot at the specified location."""
        u, v = self._uv(speed, angle)
        if 0 < speed < _BARB_BINS[0]:
            # Plot the missing barbless 1-2 knots line.
            length = self._kwargs["length"]
            pivot_points = dict(tip=0.0, middle=-length / 2.0)
            pivot = self._kwargs.get("pivot", "tip")
            offset = pivot_points[pivot]
            verts = [(0.0, offset), (0.0, length + offset)]
            rangle = math.radians(-angle)
            verts = mtransforms.Affine2D().rotate(rangle).transform(verts)
            codes = [Path.MOVETO, Path.LINETO]
            path = Path(verts, codes)
            size = length ** 2 / 4
            xy = np.array([[temperature, theta]])
            barb = PathCollection(
                [path],
                (size,),
                offsets=xy,
                transOffset=self._transform,
                **self._custom_kwargs,
            )
            barb.set_transform(mtransforms.IdentityTransform())
            self.axes.add_collection(barb)
        else:
            barb = plt.barbs(
                temperature,
                theta,
                u,
                v,
                transform=self._transform,
                **self._kwargs,
            )
        return barb

    def refresh(self):
        """Refresh the plot with the barbs."""
        if self.barbs is not None:
            xlim = self.axes.get_xlim()
            ylim = self.axes.get_ylim()
            y = np.linspace(*ylim)[::-1]
            xdelta = xlim[1] - xlim[0]
            x = np.ones(y.size) * (xlim[1] - (xdelta * self._gutter))
            xy = np.column_stack((x, y))
            points = self.axes.tephigram_inverse.transform(xy)
            temperature, theta = points[:, 0], points[:, 1]
            pressure, _ = transforms.convert_Tt2pT(temperature, theta)
            min_pressure, max_pressure = np.min(pressure), np.max(pressure)
            func = interp1d(pressure, temperature)
            for i, (speed, angle, pressure, barb) in enumerate(self.barbs):
                if min_pressure < pressure < max_pressure:
                    p2T = func(pressure)
                    temperature, theta = transforms.convert_pT2Tt(
                        pressure, p2T
                    )
                    if barb is None:
                        self.barbs[i]["barb"] = self._make_barb(
                            temperature, theta, speed, angle
                        )
                    else:
                        barb.set_offsets(np.array([[temperature, theta]]))
                        barb.set_visible(True)
                else:
                    if barb is not None:
                        barb.set_visible(False)

    def plot(self, barbs, **kwargs):
        """
        Plot the sequence of barbs.

        Args:

        * barbs:
            Sequence of speed, direction and pressure value triples for
            each barb. Where speed is measured in units of knots, direction
            in units of degrees (clockwise from north), and pressure must
            be in units of mb or hPa.

        Kwargs:

        * gutter:
            Proportion offset from the right hand side axis to plot the
            barbs. Defaults to 0.1

            Also see :func:`matplotlib.pyplot.barbs`

        """
        self._gutter = kwargs.pop("gutter", _BARB_GUTTER)
        self._kwargs = dict(length=7, zorder=10)
        self._kwargs.update(kwargs)
        self._custom_kwargs = dict(
            color=None, linewidth=1.5, zorder=self._kwargs["zorder"]
        )
        for key, values in self._custom.items():
            common = set(values).intersection(kwargs)
            if common:
                self._custom_kwargs[key] = kwargs[sorted(common)[0]]
        if hasattr(barbs, "__next__"):
            barbs = list(barbs)
        barbs = np.asarray(barbs)
        if barbs.ndim != 2 or barbs.shape[-1] != 3:
            msg = (
                "The barbs require to be a sequence of wind speed, "
                "wind direction and pressure value triples."
            )
            raise ValueError(msg)
        self.barbs = np.empty(barbs.shape[0], dtype=_BARB_DTYPE)
        for i, barb in enumerate(barbs):
            self.barbs[i] = tuple(barb) + (None,)
        self.refresh()


class Profile(object):
    """Generate an environmental lapse rate profile."""

    def __init__(self, data, axes):
        """
        Create an environmental lapse rate profile from the sequence of
        pressure and temperature point data.

        Args:

        * data:
            Sequence of pressure and temperature points defining the
            environmental lapse rate.

        * axes:
            The axes on which to plot the profile.

        """
        if hasattr(data, "__next__"):
            data = list(data)
        self.data = np.asarray(data)
        if self.data.ndim != 2 or self.data.shape[-1] != 2:
            msg = (
                "The environment profile data requires to be a sequence "
                "of pressure, temperature value pairs."
            )
            raise ValueError(msg)
        self.axes = axes
        self._transform = axes.tephigram_transform + axes.transData
        self.pressure = self.data[:, 0]
        self.temperature = self.data[:, 1]
        _, self.theta = transforms.convert_pT2Tt(
            self.pressure, self.temperature
        )
        self.line = None
        self._barbs = Barbs(axes)

    def plot(self, **kwargs):
        """
        Plot the environmental lapse rate profile.

        Kwargs:

            See :func:`matplotlib.pyplot.plot`.

        Returns:
            The profile :class:`matplotlib.lines.Line2D`

        """
        if self.line is not None and self.line in self.axes.lines:
            self.axes.lines.remove(self.line)

        if "zorder" not in kwargs:
            kwargs["zorder"] = 10

        (self.line,) = self.axes.plot(
            self.temperature, self.theta, transform=self._transform, **kwargs
        )
        return self.line

    def refresh(self):
        """Refresh the plot with the profile and any associated barbs."""
        self._barbs.refresh()

    def barbs(self, barbs, **kwargs):
        """
        Plot the sequence of barbs associated with this profile.

        Args:

        * barbs:
            Sequence of speed, direction and pressure value triples for
            each barb. Where speed is measured in units of knots, direction
            in units of degrees (clockwise from north), and pressure must
            be in units of mb or hPa.

        Kwargs:

            See :func:`matplotlib.pyplot.barbs`

        """
        colors = ["color", "barbcolor", "edgecolor", "facecolor"]
        if not set(colors).intersection(kwargs):
            kwargs["color"] = self.line.get_color()
        self._barbs.plot(barbs, **kwargs)



# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides testing capabilities and customisations specific to tephi.

.. note:: This module needs to control the matplotlib backend, so it
          **must** be imported before ``matplotlib.pyplot``.

By default, this module sets the matplotlib backend to "agg". But when
this module is imported it checks ``sys.argv`` for the flag "-d". If
found, it is removed from ``sys.argv`` and the matplotlib backend is
switched to "tkagg" to allow the interactive visual inspection of
graphical test results.

"""
import codecs
import collections
import io
import json
import os
import sys
import unittest

import filelock
import matplotlib
import numpy as np
import requests

from tephi import DATA_DIR


#: Basepath for test data.
_DATA_PATH = DATA_DIR

#: Basepath for test results.
_RESULT_PATH = os.path.join(os.path.dirname(__file__), "results")

#: Default perceptual hash size.
_HASH_SIZE = 16

#: Default maximum perceptual hash hamming distance.
_HAMMING_DISTANCE = 2

# Whether to display matplotlib output to the screen.
_DISPLAY_FIGURES = False

# Test images URL.
BASE_URL = "https://scitools.github.io/test-tephi-imagehash/images"


try:
    # Added a timeout to stop the call to requests.get hanging when running
    # on a platform which has restricted/no internet access.
    requests.get("https://github.com/SciTools/tephi", timeout=5.0)
    INET_AVAILABLE = True
except requests.exceptions.ConnectionError:
    INET_AVAILABLE = False


skip_inet = unittest.skipIf(
    not INET_AVAILABLE,
    ('Test(s) require an "internet connection", which is not available.'),
)


if "-d" in sys.argv:
    sys.argv.remove("-d")
    matplotlib.use("tkagg")
    _DISPLAY_FIGURES = True
else:
    matplotlib.use("agg")

# Imported now so that matplotlib.use can work
import matplotlib.pyplot as plt


def get_data_path(relative_path):
    """
    Returns the absolute path to a data file when given the relative path
    as a string, or sequence of strings.

    """
    if isinstance(relative_path, (list, tuple)):
        relative_path = os.path.join(*relative_path)
    return os.path.abspath(os.path.join(_DATA_PATH, relative_path))


def get_result_path(relative_path):
    """
    Returns the absolute path to a result file when given the relative path
    as a string, or sequence of strings.

    """
    if isinstance(relative_path, (list, tuple)):
        relative_path = os.path.join(*relative_path)
    return os.path.abspath(os.path.join(_RESULT_PATH, relative_path))


class TephiTest(unittest.TestCase):
    """
    A subclass of unittest.TestCase which provides testing functionality
    specific to tephi.

    """

    _assertion_counts = collections.defaultdict(int)

    def _unique_id(self):
        """
        Returns the unique ID for the current assertion.

        The ID is composed of two parts: a unique ID for the current test
        (which is itself composed of the module, class, and test names), and
        a sequential counter (specific to the current test) that is incremented
        on each call.

        For example, calls from a "test_tx" routine followed by a "test_ty"
        routine might result in::
            test_plot.TestContourf.test_tx.0
            test_plot.TestContourf.test_tx.1
            test_plot.TestContourf.test_tx.2
            test_plot.TestContourf.test_ty.0

        """
        # Obtain a consistent ID for the current test.

        # NB. unittest.TestCase.id() returns different values depending on
        # whether the test has been run explicitly, or via test discovery.
        # For example:
        #   python tests/test_brand.py
        #       => '__main__.TestBranding.test_combo'
        #   python -m unittest discover
        #       => 'tephi.tests.test_brand.TestBranding.test_combo'
        bits = self.id().split(".")[-3:]
        if bits[0] == "__main__":
            file_name = os.path.basename(sys.modules["__main__"].__file__)
            bits[0] = os.path.splitext(file_name)[0]
        test_id = ".".join(bits)

        # Derive the sequential assertion ID within the test
        assertion_id = self._assertion_counts[test_id]
        self._assertion_counts[test_id] += 1

        return "{}.{}".format(test_id, assertion_id)

    def assertArrayEqual(self, a, b):
        return np.testing.assert_array_equal(a, b)

    def assertArrayAlmostEqual(self, a, b, *args, **kwargs):
        return np.testing.assert_array_almost_equal(a, b, *args, **kwargs)


class GraphicsTest(TephiTest):
    def tearDown(self):
        # If a plotting test bombs out it can leave the current figure in an
        # odd state, so we make sure it's been disposed of.
        plt.close()

    def check_graphic(self):
        """
        Check the hash of the current matplotlib figure matches the expected
        image hash for the current graphic test.

        To create missing image test results, set the TEPHI_TEST_CREATE_MISSING
        environment variable before running the tests. This will result in new
        and appropriately "<hash>.png" image files being generated in the image
        output directory, and the imagerepo.json file being updated.

        """
        import imagehash
        from PIL import Image

        dev_mode = os.environ.get("TEPHI_TEST_CREATE_MISSING")
        unique_id = self._unique_id()
        repo_fname = os.path.join(_RESULT_PATH, "imagerepo.json")
        repo = {}
        if os.path.isfile(repo_fname):
            with open(repo_fname, "rb") as fi:
                repo = json.load(codecs.getreader("utf-8")(fi))

        try:
            #: The path where the images generated by the tests should go.
            image_output_directory = os.path.join(
                os.path.dirname(__file__), "result_image_comparison"
            )
            if not os.access(image_output_directory, os.W_OK):
                if not os.access(os.getcwd(), os.W_OK):
                    raise IOError(
                        "Write access to a local disk is required "
                        "to run image tests.  Run the tests from a "
                        "current working directory you have write "
                        "access to to avoid this issue."
                    )
                else:
                    image_output_directory = os.path.join(
                        os.getcwd(), "tephi_image_test_output"
                    )
            result_fname = os.path.join(
                image_output_directory, "result-" + unique_id + ".png"
            )

            if not os.path.isdir(image_output_directory):
                # Handle race-condition where the directories are
                # created sometime between the check above and the
                # creation attempt below.
                try:
                    os.makedirs(image_output_directory)
                except OSError as err:
                    # Don't care about "File exists"
                    if err.errno != 17:
                        raise

            def _create_missing():
                fname = f"{phash}.png"
                uri = os.path.join(BASE_URL, fname)
                hash_fname = os.path.join(image_output_directory, fname)
                uris = repo.setdefault(unique_id, [])
                uris.append(uri)
                print(f"Creating image file: {hash_fname}")
                figure.savefig(hash_fname)
                msg = "Creating imagerepo entry: {} -> {}"
                print(msg.format(unique_id, uri))
                lock = filelock.FileLock(
                    os.path.join(_RESULT_PATH, "imagerepo.lock")
                )
                # The imagerepo.json file is a critical resource, so ensure
                # thread safe read/write behaviour via platform independent
                # file locking.
                with lock.acquire(timeout=600):
                    with open(repo_fname, "wb") as fo:
                        json.dump(
                            repo,
                            codecs.getwriter("utf-8")(fo),
                            indent=4,
                            sort_keys=True,
                        )

            # Calculate the test result perceptual image hash.
            buffer = io.BytesIO()
            figure = plt.gcf()
            figure.savefig(buffer, format="png")
            buffer.seek(0)
            phash = imagehash.phash(Image.open(buffer), hash_size=_HASH_SIZE)

            if unique_id not in repo:
                if dev_mode:
                    _create_missing()
                else:
                    figure.savefig(result_fname)
                    emsg = "Missing image test result: {}."
                    raise AssertionError(emsg.format(unique_id))
            else:
                uris = repo[unique_id]
                # Extract the hex basename strings from the uris.
                hexes = [
                    os.path.splitext(os.path.basename(uri))[0] for uri in uris
                ]
                # Create the expected perceptual image hashes from the uris.
                to_hash = imagehash.hex_to_hash
                expected = [to_hash(uri_hex) for uri_hex in hexes]

                # Calculate hamming distance vector for the result hash.
                distances = [e - phash for e in expected]

                if np.all([hd > _HAMMING_DISTANCE for hd in distances]):
                    if dev_mode:
                        _create_missing()
                    else:
                        figure.savefig(result_fname)
                        msg = (
                            "Bad phash {} with hamming distance {} "
                            "for test {}."
                        )
                        msg = msg.format(phash, distances, unique_id)
                        if _DISPLAY_FIGURES:
                            emsg = "Image comparison would have failed: {}"
                            print(emsg.format(msg))
                        else:
                            emsg = "Image comparison failed: {}"
                            raise AssertionError(emsg.format(msg))

            if _DISPLAY_FIGURES:
                plt.show()

        finally:
            plt.close()



# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tephigram transform support.

"""
from matplotlib.transforms import Transform
import numpy as np

from ._constants import CONST_K, CONST_KELVIN, CONST_L, CONST_MA, CONST_RV


#
# Reference: http://www-nwp/~hadaa/tephigram/tephi_plot.html
#


def convert_Tt2pT(temperature, theta):
    """
    Transform temperature and potential temperature into
    pressure and temperature.

    Args:

    * temperature:
        Temperature in degC.

    * theta:
        Potential temperature in degC.

    Returns:
        Tuple of pressure, in mb or hPa, and temperature, in degC.

    """
    temperature, theta = np.asarray(temperature), np.asarray(theta)

    # Convert temperature and theta from degC to kelvin.
    kelvin = temperature + CONST_KELVIN
    theta = theta + CONST_KELVIN

    # Calculate the associated pressure given the temperature and
    # potential temperature.
    pressure = 1000.0 * np.power(kelvin / theta, 1 / CONST_K)

    return pressure, temperature


def convert_pT2Tt(pressure, temperature):
    """
    Transform pressure and temperature into temperature and
    potential temperature.

    Args:

    * pressure:
        Pressure in mb or hPa.

    * temperature:
        Temperature in degC.

    Returns:
        Tuple of temperature, in degC, and potential temperature, in degC.

    """
    pressure, temperature = np.asarray(pressure), np.asarray(temperature)

    # Convert temperature from degC to kelvin.
    kelvin = temperature + CONST_KELVIN

    # Calculate the potential temperature given the pressure and temperature.
    theta = kelvin * ((1000.0 / pressure) ** CONST_K)

    # Convert potential temperature from kelvin to degC.
    return temperature, theta - CONST_KELVIN


def convert_pt2pT(pressure, theta):
    """
    Transform pressure and potential temperature into pressure and temperature.

    Args:

    * pressure:
        Pressure in mb or hPa.

    * theta:
        Potential temperature in degC.

    * Returns:
        Tuple of pressure, in mb or hPa, and temperature, in degC.

    """
    pressure, theta = np.asarray(pressure), np.asarray(theta)

    # Convert potential temperature from degC to kelvin.
    theta = theta + CONST_KELVIN

    # Calculate the temperature given the pressure and
    # potential temperature.
    kelvin = theta * (pressure ** CONST_K) / (1000.0 ** CONST_K)

    # Convert temperature from kelvin to degC.
    return pressure, kelvin - CONST_KELVIN


def convert_Tt2xy(temperature, theta):
    """
    Transform temperature and potential temperature to native display
    coordinates.

    Args:

    * temperature:
        Temperature in degC.

    * theta:
        Potential temperature in degC.

    Returns:
        Native display x and y coordinates.

    """
    temperature, theta = np.asarray(temperature), np.asarray(theta)

    # Convert potential temperature from degC to kelvin.
    theta = theta + CONST_KELVIN
    theta = np.clip(theta, 1, 1e10)

    phi = np.log(theta)

    x_data = phi * CONST_MA + temperature
    y_data = phi * CONST_MA - temperature

    return x_data, y_data


def convert_xy2Tt(x_data, y_data):
    """
    Transform native display coordinates to temperature and
    potential temperature.

    Args:

    * x_data:
        Native display x-coordinate/s.

    * y_data:
        Native display y-coordinate/s.

    Returns:
        Temperature, in degC, and potential temperature, in degC.

    """
    x_data, y_data = np.asarray(x_data), np.asarray(y_data)

    phi = (x_data + y_data) / (2 * CONST_MA)
    temperature = (x_data - y_data) / 2.0

    theta = np.exp(phi) - CONST_KELVIN

    return temperature, theta


def convert_pw2T(pressure, mixing_ratio):
    """
    Transform pressure and mixing ratios to temperature.

    Args:

    * pressure:
        Pressure in mb in hPa.

    * mixing_ratio:
        Dimensionless mixing ratios.

    Returns:
        Temperature in degC.

    """
    pressure = np.array(pressure)

    # Calculate the dew-point.
    vapp = pressure * (8.0 / 5.0) * (mixing_ratio / 1000.0)
    temp = 1.0 / (
        (1.0 / CONST_KELVIN) - ((CONST_RV / CONST_L) * np.log(vapp / 6.11))
    )

    return temp - CONST_KELVIN


class TephiTransform(Transform):
    """
    Tephigram transformation to convert from temperature and
    potential temperature to native plotting device coordinates.

    """

    input_dims = 2
    output_dims = 2
    is_separable = False
    has_inverse = True

    def transform_non_affine(self, values):
        """
        Transform from tephigram temperature and potential temperature
        to native plotting device coordinates.

        Args:

        * values:
            Values to be transformed, with shape (N, 2).

        """
        return np.concatenate(
            convert_Tt2xy(values[:, 0:1], values[:, 1:2]), axis=1
        )

    def inverted(self):
        """Return the inverse transformation."""
        return TephiTransformInverted()


class TephiTransformInverted(Transform):
    """
    Tephigram inverse transformation to convert from native
    plotting device coordinates to tephigram temperature and
    potential temperature.

    """

    input_dims = 2
    output_dims = 2
    is_separable = False
    has_inverse = True

    def transform_non_affine(self, values):
        """
        Transform from native plotting display coordinates to tephigram
        temperature and potential temperature.

        Args:

        * values:
           Values to be transformed, with shape (N, 2).

        """
        return np.concatenate(
            convert_xy2Tt(values[:, 0:1], values[:, 1:2]), axis=1
        )

    def inverted(self):
        """Return the inverse transformation."""
        return TephiTransform()



# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests the tephigram plotting capability provided by tephi.

"""
# Import tephi test package first so that some things can be initialised
# before importing anything else.
import tephi.tests as tests

import unittest

import numpy as np

import tephi
from tephi import Tephigram


def _load_result(filename):
    with np.load(tephi.tests.get_result_path(filename)) as f:
        result = f["arr_0"]
    return result


_expected_dews = _load_result("dews.npz")
_expected_temps = _load_result("temps.npz")
_expected_barbs = _load_result("barbs.npz")


class TestTephigramLoadTxt(tests.TephiTest):
    def setUp(self):
        self.filename_dews = tephi.tests.get_data_path("dews.txt")
        self.filename_temps = tephi.tests.get_data_path("temps.txt")
        self.filename_barbs = tephi.tests.get_data_path("barbs.txt")
        self.filename_comma = tephi.tests.get_data_path("comma_sep.txt")

    def test_is_not_file(self):
        with self.assertRaises(OSError):
            tephi.loadtxt("wibble")

    def test_load_data_no_column_names(self):
        dews = tephi.loadtxt(self.filename_dews)
        self.assertEqual(dews._fields, ("pressure", "temperature"))
        self.assertArrayEqual(dews.pressure, _expected_dews[0])
        self.assertArrayEqual(dews, _expected_dews)

    def test_load_data_with_column_names(self):
        # Column titles test all valid namedtuple characters (alphanumeric, _).
        columns = ("pressure", "dewpoint2", "wind_speed", "WindDirection")
        barbs = tephi.loadtxt(self.filename_barbs, column_titles=columns)
        self.assertEqual(barbs._fields, columns)
        self.assertArrayEqual(barbs.wind_speed, _expected_barbs[2])
        self.assertArrayEqual(barbs, _expected_barbs)

    def test_load_multiple_files_same_column_names(self):
        columns = ("foo", "bar")
        dews, temps = tephi.loadtxt(
            self.filename_dews, self.filename_temps, column_titles=columns
        )
        self.assertEqual(dews._fields, columns)
        self.assertEqual(temps._fields, columns)

    def test_load_data_too_many_column_iterables(self):
        columns = [
            ("pressure", "dewpoint"),
            ("pressure", "wind_speed", "wind_direction"),
        ]
        with self.assertRaises(ValueError):
            tephi.loadtxt(self.filename_dews, column_titles=columns)

    def test_number_of_columns_and_titles_not_equal(self):
        columns = ("pressure", "dewpoint", "wind_speed")
        with self.assertRaises(TypeError):
            tephi.loadtxt(self.filename_barbs, column_titles=columns)

    def test_invalid_column_titles(self):
        columns = ("pres-sure", "dew+point", 5)
        with self.assertRaises(ValueError):
            tephi.loadtxt(self.filename_dews, column_titles=columns)

    def test_non_iterable_column_title(self):
        # For the case of column titles, strings are considered non-iterable.
        columns = "pressure"
        with self.assertRaises(TypeError):
            tephi.loadtxt(self.filename_dews, column_titles=columns)

    def test_delimiter(self):
        columns = ("pressure", "temperature", "wind_direction", "wind_speed")
        data = tephi.loadtxt(
            self.filename_comma, column_titles=columns, delimiter=","
        )
        self.assertEqual(data.pressure.shape, (2,))

    def test_dtype(self):
        dews = tephi.loadtxt(self.filename_dews, dtype="i4")
        self.assertIsInstance(dews.pressure[0], np.int32)
        self.assertIsInstance(dews.temperature[0], np.int32)


class TestTephigramPlot(tests.GraphicsTest):
    def setUp(self):
        self.dews = _expected_dews.T
        self.temps = _expected_temps.T

    def test_plot_dews(self):
        tpg = Tephigram()
        tpg.plot(self.dews)
        self.check_graphic()

    def test_plot_temps(self):
        tpg = Tephigram()
        tpg.plot(self.temps)
        self.check_graphic()

    def test_plot_dews_temps(self):
        tpg = Tephigram()
        tpg.plot(self.dews)
        tpg.plot(self.temps)
        self.check_graphic()

    def test_plot_dews_label(self):
        tpg = Tephigram()
        tpg.plot(self.dews, label="Dew-point temperature")
        self.check_graphic()

    def test_plot_temps_label(self):
        tpg = Tephigram()
        tpg.plot(self.temps, label="Dry-bulb temperature")
        self.check_graphic()

    def test_plot_dews_custom(self):
        tpg = Tephigram()
        tpg.plot(
            self.dews,
            label="Dew-point temperature",
            linewidth=2,
            color="blue",
            marker="s",
        )
        self.check_graphic()

    def test_plot_temps_custom(self):
        tpg = Tephigram()
        tpg.plot(
            self.temps,
            label="Dry-bulb temperature",
            linewidth=2,
            color="red",
            marker="o",
        )
        self.check_graphic()

    def test_plot_dews_temps_custom(self):
        tpg = Tephigram()
        tpg.plot(
            self.dews,
            label="Dew-point temperature",
            linewidth=2,
            color="blue",
            marker="s",
        )
        tpg.plot(
            self.temps,
            label="Dry-bulb temperature",
            linewidth=2,
            color="red",
            marker="o",
        )
        self.check_graphic()

    def test_plot_dews_locator_isotherm_numeric(self):
        tpg = Tephigram(isotherm_locator=10)
        tpg.plot(self.dews)
        self.check_graphic()

    def test_plot_dews_locator_isotherm_object(self):
        tpg = Tephigram(isotherm_locator=tephi.Locator(10))
        tpg.plot(self.dews)
        self.check_graphic()

    def test_plot_dews_locator_adiabat_numeric(self):
        tpg = Tephigram(dry_adiabat_locator=10)
        tpg.plot(self.dews)
        self.check_graphic()

    def test_plot_dews_locator_adiabat_object(self):
        tpg = Tephigram(dry_adiabat_locator=tephi.Locator(10))
        tpg.plot(self.dews)
        self.check_graphic()

    def test_plot_dews_locator_numeric(self):
        tpg = Tephigram(isotherm_locator=10, dry_adiabat_locator=10)
        tpg.plot(self.dews)
        self.check_graphic()

    def test_plot_dews_locator_object(self):
        locator = tephi.Locator(10)
        tpg = Tephigram(isotherm_locator=locator, dry_adiabat_locator=locator)
        tpg.plot(self.dews)
        self.check_graphic()

    def test_plot_anchor(self):
        tpg = Tephigram(anchor=[(1000, 0), (300, 0)])
        tpg.plot(self.dews)
        self.check_graphic()


class TestTephigramBarbs(tests.GraphicsTest):
    def setUp(self):
        self.dews = _expected_dews.T
        self.temps = _expected_temps.T
        magnitude = np.hstack(([0], np.arange(20) * 5 + 2, [102]))
        self.barbs = [(m, 45, 1000 - i * 35) for i, m in enumerate(magnitude)]

    def test_rotate(self):
        tpg = Tephigram()
        profile = tpg.plot(self.temps)
        profile.barbs(
            [
                (0, 0, 900),
                (1, 30, 850),
                (5, 60, 800),
                (10, 90, 750),
                (15, 120, 700),
                (20, 150, 650),
                (25, 180, 600),
                (30, 210, 550),
                (35, 240, 500),
                (40, 270, 450),
                (45, 300, 400),
                (50, 330, 350),
                (55, 360, 300),
            ],
            zorder=10,
        )
        self.check_graphic()

    def test_barbs(self):
        tpg = Tephigram()
        profile = tpg.plot(self.temps)
        profile.barbs(self.barbs, zorder=10)
        self.check_graphic()

    def test_barbs_from_file(self):
        tpg = Tephigram()
        dews = _expected_barbs.T[:, :2]
        barbs = np.column_stack(
            (_expected_barbs[2], _expected_barbs[3], _expected_barbs[0])
        )
        profile = tpg.plot(dews)
        profile.barbs(barbs, zorder=10)
        self.check_graphic()

    def test_gutter(self):
        tpg = Tephigram()
        profile = tpg.plot(self.temps)
        profile.barbs(self.barbs, gutter=0.5, zorder=10)
        self.check_graphic()

    def test_length(self):
        tpg = Tephigram()
        profile = tpg.plot(self.temps)
        profile.barbs(self.barbs, gutter=0.9, length=10, zorder=10)
        self.check_graphic()

    def test_color(self):
        tpg = Tephigram()
        profile = tpg.plot(self.temps)
        profile.barbs(self.barbs, color="green", zorder=10)
        self.check_graphic()

    def test_pivot(self):
        tpg = Tephigram()
        tprofile = tpg.plot(self.temps)
        tprofile.barbs(self.barbs, gutter=0.2, pivot="tip", length=8)
        dprofile = tpg.plot(self.dews)
        dprofile.barbs(self.barbs, gutter=0.3, pivot="middle", length=8)
        self.check_graphic()


if __name__ == "__main__":
    unittest.main()



# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test that imagerepo.json reconciles with the images registered in the
test-tephi-imagehash repository.

"""
# Import tephi tests first so that some things can be initialised before
# importing anything else
import tephi.tests as tests

import codecs
import itertools
import json
import os
import unittest

import requests


IMAGE_MANIFEST = (
    "https://raw.githubusercontent.com/SciTools/"
    "test-tephi-imagehash/gh-pages/image_manifest.txt"
)


@tests.skip_inet
class TestImageRepoJSON(tests.TephiTest):
    def test(self):
        response = requests.get(IMAGE_MANIFEST)

        emsg = 'Failed to download "image_manifest.txt"'
        self.assertEqual(response.status_code, requests.codes.ok, msg=emsg)

        image_manifest = response.content.decode("utf-8")
        image_manifest = [line.strip() for line in image_manifest.split("\n")]
        image_manifest_uris = set(
            os.path.join(tests.BASE_URL, fname) for fname in image_manifest
        )

        imagerepo_fname = os.path.join(
            os.path.dirname(__file__), "results", "imagerepo.json"
        )
        with open(imagerepo_fname, "rb") as fi:
            imagerepo = json.load(codecs.getreader("utf-8")(fi))

        # "imagerepo" maps key: list-of-uris. Put all the uris in one big set.
        tests_uris = set(itertools.chain.from_iterable(imagerepo.values()))

        missing = list(tests_uris - image_manifest_uris)
        count = len(missing)
        if count:
            emsg = (
                '"imagerepo.json" references {} image URIs that are not '
                'listed in "{}":\n\t'
            )
            emsg = emsg.format(count, IMAGE_MANIFEST)
            emsg += "\t".join(uri for uri in missing)
            # Always fails when we get here: report the problem.
            self.assertEqual(count, 0, msg=emsg)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

from setuptools import setup


if __name__ == "__main__":
    setup()



# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tephigram transform and isopleth constants.

"""

# TODO: Discover the meaning of the magic constant numbers.

CONST_CP = 1.01e3
CONST_K = 0.286
CONST_KELVIN = 273.15  # celsius to kelvin offset.
CONST_L = 2.5e6
CONST_MA = 300.0
CONST_RD = 287.0
CONST_RV = 461.0
